# coding=utf-8
"""OpenStudio ConstructionSet translator."""
from __future__ import division

from honeybee.typing import clean_ep_string
from honeybee_energy.construction.windowshade import WindowConstructionShade
from honeybee_energy.construction.dynamic import WindowConstructionDynamic
from honeybee_energy.constructionset import ConstructionSet

from honeybee_openstudio.openstudio import OSDefaultConstructionSet, \
    OSDefaultSurfaceConstructions, OSDefaultSubSurfaceConstructions
from honeybee_openstudio.construction import shade_construction_from_openstudio


def _assign_construction_to_subset(construction, os_constr_subset, face_type, os_model):
    """Assign a Honeybee construction object to an OpenStudio sub-set.

    Args:
        construction: The honeybee-energy construction object assigned to the
            ConstructionsSet (this can be None).
        os_constr_subset: The OpenStudio DefaultSurfaceConstructions object
            to which the construction will be added.
        face_type: Text for the type of Face to which the construction will be
            added. Must be either Wall, Floor or RoofCeiling.
        os_model: The OpenStudio Model.
    """
    if construction is not None:
        construction_ref = os_model.getConstructionByName(construction.identifier)
        if construction_ref.is_initialized():
            os_construction = construction_ref.get()
            if face_type == 'Wall':
                os_constr_subset.setWallConstruction(os_construction)
            elif face_type == 'Floor':
                os_constr_subset.setFloorConstruction(os_construction)
            else:
                os_constr_subset.setRoofCeilingConstruction(os_construction)


def _glazing_construction(construction, os_model):
    """Get an OpenStudio window construction with a check for dynamic constructions."""
    if construction is None:
        return None
    elif isinstance(construction, WindowConstructionShade):
        construction_id = construction.window_construction.identifier
    elif isinstance(construction, WindowConstructionDynamic):
        construction_id = '{}State0'.format(construction.constructions[0].identifier)
    else:
        construction_id = construction.identifier
    constr_ref = os_model.getConstructionByName(construction_id)
    if constr_ref.is_initialized():
        os_construction = constr_ref.get()
        return os_construction


def construction_set_to_openstudio(construction_set, os_model):
    """Convert Honeybee ConstructionSet to OpenStudio DefaultConstructionSet."""
    # create the construction set object
    os_constr_set = OSDefaultConstructionSet(os_model)
    os_constr_set.setName(construction_set.identifier)
    if construction_set._display_name is not None:
        os_constr_set.setDisplayName(construction_set.display_name)

    int_surf_const = OSDefaultSurfaceConstructions(os_model)
    ext_surf_const = OSDefaultSurfaceConstructions(os_model)
    grnd_surf_const = OSDefaultSurfaceConstructions(os_model)
    int_subsurf_const = OSDefaultSubSurfaceConstructions(os_model)
    ext_subsurf_const = OSDefaultSubSurfaceConstructions(os_model)

    os_constr_set.setDefaultInteriorSurfaceConstructions(int_surf_const)
    os_constr_set.setDefaultExteriorSurfaceConstructions(ext_surf_const)
    os_constr_set.setDefaultGroundContactSurfaceConstructions(grnd_surf_const)
    os_constr_set.setDefaultInteriorSubSurfaceConstructions(int_subsurf_const)
    os_constr_set.setDefaultExteriorSubSurfaceConstructions(ext_subsurf_const)

    # determine the frame type for measure tags
    frame_type = 'Metal Framing with Thermal Break' \
        if 'WoodFramed' in construction_set.identifier else 'Non-Metal Framing'

    # assign the constructions in the wall set
    int_con = construction_set.wall_set._interior_construction
    if int_con is not None:
        int_wall_ref = os_model.getConstructionByName(int_con.identifier)
        if int_wall_ref.is_initialized():
            interior_wall = int_wall_ref.get()
            int_surf_const.setWallConstruction(interior_wall)
            os_constr_set.setAdiabaticSurfaceConstruction(interior_wall)
    ext_con = construction_set.wall_set._exterior_construction
    _assign_construction_to_subset(ext_con, ext_surf_const, 'Wall', os_model)
    ground_con = construction_set.wall_set._ground_construction
    _assign_construction_to_subset(ground_con, grnd_surf_const, 'Wall', os_model)

    # assign the constructions in the floor set
    int_con = construction_set.floor_set._interior_construction
    _assign_construction_to_subset(int_con, int_surf_const, 'Floor', os_model)
    ext_con = construction_set.floor_set._exterior_construction
    _assign_construction_to_subset(ext_con, ext_surf_const, 'Floor', os_model)
    ground_con = construction_set.floor_set._ground_construction
    _assign_construction_to_subset(ground_con, grnd_surf_const, 'Floor', os_model)

    # assign the constructions in the roof ceiling set
    int_con = construction_set.roof_ceiling_set._interior_construction
    _assign_construction_to_subset(int_con, int_surf_const, 'RoofCeiling', os_model)
    ext_con = construction_set.roof_ceiling_set._exterior_construction
    _assign_construction_to_subset(ext_con, ext_surf_const, 'RoofCeiling', os_model)
    ground_con = construction_set.roof_ceiling_set._ground_construction
    _assign_construction_to_subset(ground_con, grnd_surf_const, 'RoofCeiling', os_model)

    # assign the constructions in the aperture set
    int_ap_con = construction_set.aperture_set._interior_construction
    int_ap_con = _glazing_construction(int_ap_con, os_model)
    if int_ap_con is not None:
        int_subsurf_const.setFixedWindowConstruction(int_ap_con)
        int_subsurf_const.setOperableWindowConstruction(int_ap_con)
        int_subsurf_const.setSkylightConstruction(int_ap_con)
    win_ap_con = construction_set.aperture_set._window_construction
    win_ap_con = _glazing_construction(win_ap_con, os_model)
    if win_ap_con is not None:
        ext_subsurf_const.setFixedWindowConstruction(win_ap_con)
        std_info = win_ap_con.standardsInformation()
        std_info.setFenestrationType('Fixed Window')
        std_info.setFenestrationFrameType(frame_type)
        std_info.setIntendedSurfaceType('ExteriorWindow')
    sky_ap_con = construction_set.aperture_set._skylight_construction
    sky_ap_con = _glazing_construction(sky_ap_con, os_model)
    if sky_ap_con is not None:
        ext_subsurf_const.setSkylightConstruction(sky_ap_con)
        std_info = sky_ap_con.standardsInformation()
        std_info.setFenestrationType('Fixed Window')
        std_info.setFenestrationFrameType(frame_type)
        if not std_info.intendedSurfaceType().is_initialized():
            std_info.setIntendedSurfaceType('Skylight')
    op_ap_con = construction_set.aperture_set._operable_construction
    op_ap_con = _glazing_construction(op_ap_con, os_model)
    if op_ap_con is not None:
        ext_subsurf_const.setOperableWindowConstruction(op_ap_con)
        std_info = op_ap_con.standardsInformation()
        std_info.setFenestrationFrameType(frame_type)
        std_info.setIntendedSurfaceType('ExteriorWindow')
        if not std_info.intendedSurfaceType().is_initialized():
            std_info.setFenestrationType('Operable Window')

    # assign the constructions in the door set
    int_dr_con = construction_set.door_set._interior_construction
    if int_dr_con is not None:
        int_door_ref = os_model.getConstructionByName(int_dr_con.identifier)
        if int_door_ref.is_initialized():
            interior_door = int_door_ref.get()
            int_subsurf_const.setDoorConstruction(interior_door)
            int_subsurf_const.setOverheadDoorConstruction(interior_door)
    ext_dr_con = construction_set.door_set._exterior_construction
    if ext_dr_con is not None:
        ext_door_ref = os_model.getConstructionByName(ext_dr_con.identifier)
        if ext_door_ref.is_initialized():
            exterior_door = ext_door_ref.get()
            ext_subsurf_const.setDoorConstruction(exterior_door)
            std_info = exterior_door.standardsInformation()
            if not std_info.intendedSurfaceType().is_initialized():
                std_info.setIntendedSurfaceType('ExteriorDoor')
    ov_dr_con = construction_set.door_set._overhead_construction
    if ov_dr_con is not None:
        overhead_door_ref = os_model.getConstructionByName(ov_dr_con.identifier)
        if overhead_door_ref.is_initialized():
            overhead_door = overhead_door_ref.get()
            ext_subsurf_const.setOverheadDoorConstruction(overhead_door)
            std_info = overhead_door.standardsInformation()
            if not std_info.intendedSurfaceType().is_initialized():
                std_info.setIntendedSurfaceType('OverheadDoor')
    ext_glz_for_con = construction_set.door_set._exterior_glass_construction
    ext_glz_for_con = _glazing_construction(ext_glz_for_con, os_model)
    if ext_glz_for_con is not None:
        ext_subsurf_const.setGlassDoorConstruction(ext_glz_for_con)
        std_info = ext_glz_for_con.standardsInformation()
        if not std_info.fenestrationType().is_initialized():
            std_info.setFenestrationType('Glazed Door')
        std_info.setFenestrationFrameType(frame_type)
        if not std_info.intendedSurfaceType().is_initialized():
            std_info.setIntendedSurfaceType('GlassDoor')
    int_glz_for_con = construction_set.door_set._interior_glass_construction
    int_glz_for_con = _glazing_construction(int_glz_for_con, os_model)
    if int_glz_for_con is not None:
        int_subsurf_const.setGlassDoorConstruction(int_glz_for_con)

    # assign the shading construction to construction set
    shade_con = construction_set._shade_construction
    if shade_con is not None:
        shade_ref = os_model.getConstructionByName(shade_con.identifier)
        if shade_ref.is_initialized():
            shade_construction = shade_ref.get()
            os_constr_set.setSpaceShadingConstruction(shade_construction)

    return os_constr_set


def construction_set_from_openstudio(os_construction_set, constructions):
    """Convert OpenStudio DefaultConstructionSet to Honeybee ConstructionSet."""
    con_set = ConstructionSet(clean_ep_string(os_construction_set.nameString()))

    # get interior surface constructions
    if os_construction_set.defaultInteriorSurfaceConstructions().is_initialized():
        os_int_set = os_construction_set.defaultInteriorSurfaceConstructions().get()
        if os_int_set.wallConstruction().is_initialized():
            int_wall_const = os_int_set.wallConstruction().get().nameString()
            try:
                int_wall_const = constructions[clean_ep_string(int_wall_const)]
                con_set.wall_set.interior_construction = int_wall_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if os_int_set.floorConstruction().is_initialized():
            int_floor_const = os_int_set.floorConstruction().get().nameString()
            try:
                int_floor_const = constructions[clean_ep_string(int_floor_const)]
                con_set.floor_set.interior_construction = int_floor_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if os_int_set.roofCeilingConstruction().is_initialized():
            int_roof_const = os_int_set.roofCeilingConstruction().get().nameString()
            try:
                int_roof_const = constructions[clean_ep_string(int_roof_const)]
                con_set.roof_ceiling_set.interior_construction = int_roof_const
            except KeyError:
                pass  # construction that could not be re-serialized

    # get interior subsurface constructions
    if os_construction_set.defaultInteriorSubSurfaceConstructions().is_initialized():
        int_subset = os_construction_set.defaultInteriorSubSurfaceConstructions().get()
        if int_subset.fixedWindowConstruction().is_initialized():
            int_wind_const = int_subset.fixedWindowConstruction().get().nameString()
            try:
                int_wind_const = constructions[clean_ep_string(int_wind_const)]
                con_set.aperture_set.interior_construction = int_wind_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if int_subset.doorConstruction().is_initialized():
            int_door_const = int_subset.doorConstruction().get().nameString()
            try:
                int_door_const = constructions[clean_ep_string(int_door_const)]
                con_set.door_set.interior_construction = int_door_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if int_subset.glassDoorConstruction().is_initialized():
            int_glass_door_con = int_subset.glassDoorConstruction().get().nameString()
            try:
                int_glass_door_con = constructions[clean_ep_string(int_glass_door_con)]
                con_set.door_set.interior_glass_construction = int_glass_door_con
            except KeyError:
                pass  # construction that could not be re-serialized

    # get exterior surface constructions
    if os_construction_set.defaultExteriorSurfaceConstructions().is_initialized():
        os_ext_set = os_construction_set.defaultExteriorSurfaceConstructions().get()
        if os_ext_set.wallConstruction().is_initialized():
            ext_wall_const = os_ext_set.wallConstruction().get().nameString()
            try:
                ext_wall_const = constructions[clean_ep_string(ext_wall_const)]
                con_set.wall_set.exterior_construction = ext_wall_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if os_ext_set.floorConstruction().is_initialized():
            ext_floor_const = os_ext_set.floorConstruction().get().nameString()
            try:
                ext_floor_const = constructions[clean_ep_string(ext_floor_const)]
                con_set.floor_set.exterior_construction = ext_floor_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if os_ext_set.roofCeilingConstruction().is_initialized():
            ext_roof_const = os_ext_set.roofCeilingConstruction().get().nameString()
            try:
                ext_roof_const = constructions[clean_ep_string(ext_roof_const)]
                con_set.roof_ceiling_set.exterior_construction = ext_roof_const
            except KeyError:
                pass  # construction that could not be re-serialized

    # get exterior subsurface construction
    if os_construction_set.defaultExteriorSubSurfaceConstructions().is_initialized():
        ext_subset = os_construction_set.defaultExteriorSubSurfaceConstructions().get()
        if ext_subset.fixedWindowConstruction().is_initialized():
            ext_wind_const = ext_subset.fixedWindowConstruction().get().nameString()
            try:
                ext_wind_const = constructions[clean_ep_string(ext_wind_const)]
                con_set.aperture_set.window_construction = ext_wind_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if ext_subset.operableWindowConstruction().is_initialized():
            op_wind_const = ext_subset.operableWindowConstruction().get().nameString()
            try:
                op_wind_const = constructions[clean_ep_string(op_wind_const)]
                con_set.aperture_set.operable_construction = op_wind_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if ext_subset.skylightConstruction().is_initialized():
            ext_skylight_const = ext_subset.skylightConstruction().get().nameString()
            try:
                ext_skylight_const = constructions[clean_ep_string(ext_skylight_const)]
                con_set.aperture_set.skylight_construction = ext_skylight_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if ext_subset.doorConstruction().is_initialized():
            ext_door_const = ext_subset.doorConstruction().get().nameString()
            try:
                ext_door_const = constructions[clean_ep_string(ext_door_const)]
                con_set.door_set.exterior_construction = ext_door_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if ext_subset.overheadDoorConstruction().is_initialized():
            ext_ovhd_door = ext_subset.overheadDoorConstruction().get().nameString()
            try:
                ext_ovhd_door = constructions[clean_ep_string(ext_ovhd_door)]
                con_set.door_set.overhead_construction = ext_ovhd_door
            except KeyError:
                pass  # construction that could not be re-serialized
        if ext_subset.glassDoorConstruction().is_initialized():
            ext_glz_door = ext_subset.glassDoorConstruction().get().nameString()
            try:
                ext_glz_door = constructions[clean_ep_string(ext_glz_door)]
                con_set.door_set.exterior_glass_construction = ext_glz_door
            except KeyError:
                pass  # construction that could not be re-serialized

    # assign the ground construction and other attributes
    if os_construction_set.defaultGroundContactSurfaceConstructions().is_initialized():
        os_gnd_set = os_construction_set.defaultGroundContactSurfaceConstructions().get()
        if os_gnd_set.wallConstruction().is_initialized():
            gnd_wall_const = os_gnd_set.wallConstruction().get().nameString()
            try:
                gnd_wall_const = constructions[clean_ep_string(gnd_wall_const)]
                con_set.wall_set.ground_construction = gnd_wall_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if os_gnd_set.floorConstruction().is_initialized():
            gnd_floor_const = os_gnd_set.floorConstruction().get().nameString()
            try:
                gnd_floor_const = constructions[clean_ep_string(gnd_floor_const)]
                con_set.floor_set.ground_construction = gnd_floor_const
            except KeyError:
                pass  # construction that could not be re-serialized
        if os_gnd_set.roofCeilingConstruction().is_initialized():
            gnd_roof_const = os_gnd_set.roofCeilingConstruction().get().nameString()
            try:
                gnd_roof_const = constructions[clean_ep_string(gnd_roof_const)]
                con_set.roof_ceiling_set.ground_construction = gnd_roof_const
            except KeyError:
                pass  # construction that could not be re-serialized

    # assign shade and other optional attributes
    if os_construction_set.spaceShadingConstruction().is_initialized():
        shade_con = os_construction_set.spaceShadingConstruction().get()
        const_name = '{} Shade'.format(clean_ep_string(shade_con.nameString()))
        try:
            shade_con = constructions[const_name]
        except KeyError:
            const = shade_con.to_LayeredConstruction().get()
            shade_con = shade_construction_from_openstudio(const)
            constructions[const_name] = shade_con
        con_set.shade_construction = shade_con
    if os_construction_set.displayName().is_initialized():
        con_set.display_name = os_construction_set.displayName().get()
    return con_set



# coding=utf-8
"""Module taken from OpenStudio-standards.

https://github.com/NREL/openstudio-standards/blob/master/
lib/openstudio-standards/prototypes/common/objects/Prototype.CentralAirSourceHeatPump.rb
"""
from __future__ import division

from honeybee_openstudio.openstudio import openstudio_model
from .utilities import ems_friendly_name


def create_central_air_source_heat_pump(model, hot_water_loop, name=None, cop=3.65):
    """Prototype CentralAirSourceHeatPump object using PlantComponentUserDefined.

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        hot_water_loop: [<OpenStudio::Model::PlantLoop>] a hot water loop served
            by the central air source heat pump.
        name: [String] the name of the central air source heat pump, or nil in
            which case it will be defaulted.
        cop: [Double] air source heat pump rated cop.
    """
    # create the PlantComponentUserDefined object as a proxy for the Central Air Source Heat Pump
    plant_comp = openstudio_model.PlantComponentUserDefined(model)
    if name is None:
        if hot_water_loop is None:
            name = 'Central Air Source Heat Pump'
        else:
            name = '{} Central Air Source Heat Pump'.format(hot_water_loop.nameString())

    # change equipment name for EMS validity
    plant_comp.setName(ems_friendly_name(name))

    # set plant component properties
    plant_comp.setPlantLoadingMode('MeetsLoadWithNominalCapacityHiOutLimit')
    plant_comp.setPlantLoopFlowRequestMode('NeedsFlowIfLoopOn')

    # plant design volume flow rate internal variable
    vdot_des_int_var = openstudio_model.EnergyManagementSystemInternalVariable(
        model, 'Plant Design Volume Flow Rate')
    vdot_des_int_var.setName('{}_Vdot_Des_Int_Var'.format(plant_comp.nameString()))
    vdot_des_int_var.setInternalDataIndexKeyName(str(hot_water_loop.handle()))

    # inlet temperature internal variable
    tin_int_var = openstudio_model.EnergyManagementSystemInternalVariable(
        model, 'Inlet Temperature for Plant Connection 1')
    tin_int_var.setName('{}_Tin_Int_Var'.format(plant_comp.nameString()))
    tin_int_var.setInternalDataIndexKeyName(str(plant_comp.handle()))

    # inlet mass flow rate internal variable
    mdot_int_var = openstudio_model.EnergyManagementSystemInternalVariable(
        model, 'Inlet Mass Flow Rate for Plant Connection 1')
    mdot_int_var.setName('{}_Mdot_Int_Var'.format(plant_comp.nameString()))
    mdot_int_var.setInternalDataIndexKeyName(str(plant_comp.handle()))

    # inlet specific heat internal variable
    cp_int_var = openstudio_model.EnergyManagementSystemInternalVariable(
        model, 'Inlet Specific Heat for Plant Connection 1')
    cp_int_var.setName('{}_Cp_Int_Var'.format(plant_comp.nameString()))
    cp_int_var.setInternalDataIndexKeyName(str(plant_comp.handle()))

    # inlet density internal variable
    rho_int_var = openstudio_model.EnergyManagementSystemInternalVariable(
        model, 'Inlet Density for Plant Connection 1')
    rho_int_var.setName('{}_rho_Int_Var'.format(plant_comp.nameString()))
    rho_int_var.setInternalDataIndexKeyName(str(plant_comp.handle()))

    # load request internal variable
    load_int_var = openstudio_model.EnergyManagementSystemInternalVariable(
        model, 'Load Request for Plant Connection 1')
    load_int_var.setName('{}_Load_Int_Var'.format(plant_comp.nameString()))
    load_int_var.setInternalDataIndexKeyName(str(plant_comp.handle()))

    # supply outlet node setpoint temperature sensor
    setpt_mgr_sch_sen = openstudio_model.EnergyManagementSystemSensor(
        model, 'Schedule Value')
    setpt_mgr_sch_sen.setName('{}_Setpt_Mgr_Temp_Sen'.format(plant_comp.nameString()))
    for m in hot_water_loop.supplyOutletNode().setpointManagers():
        if m.to_SetpointManagerScheduled().is_initialized():
            m_name = m.to_SetpointManagerScheduled().get().schedule().nameString()
            setpt_mgr_sch_sen.setKeyName(m_name)

    # outdoor air drybulb temperature sensor
    oa_dbt_sen = openstudio_model.EnergyManagementSystemSensor(
        model, 'Site Outdoor Air Drybulb Temperature')
    oa_dbt_sen.setName('{}_OA_DBT_Sen'.format(plant_comp.nameString()))
    oa_dbt_sen.setKeyName('Environment')

    # minimum mass flow rate actuator
    mdot_min_act = plant_comp.minimumMassFlowRateActuator().get()
    mdot_min_act.setName('{}_Mdot_Min_Act'.format(plant_comp.nameString()))

    # maximum mass flow rate actuator
    mdot_max_act = plant_comp.maximumMassFlowRateActuator().get()
    mdot_max_act.setName('{}_Mdot_Max_Act'.format(plant_comp.nameString()))

    # design flow rate actuator
    vdot_des_act = plant_comp.designVolumeFlowRateActuator().get()
    vdot_des_act.setName('{}_Vdot_Des_Act'.format(plant_comp.nameString()))

    # minimum loading capacity actuator
    cap_min_act = plant_comp.minimumLoadingCapacityActuator().get()
    cap_min_act.setName('{}_Cap_Min_Act'.format(plant_comp.nameString()))

    # maximum loading capacity actuator
    cap_max_act = plant_comp.maximumLoadingCapacityActuator().get()
    cap_max_act.setName('{}_Cap_Max_Act'.format(plant_comp.nameString()))

    # optimal loading capacity actuator
    cap_opt_act = plant_comp.optimalLoadingCapacityActuator().get()
    cap_opt_act.setName('{}_Cap_Opt_Act'.format(plant_comp.nameString()))

    # outlet temperature actuator
    tout_act = plant_comp.outletTemperatureActuator().get()
    tout_act.setName('{}_Tout_Act'.format(plant_comp.nameString()))

    # mass flow rate actuator
    mdot_req_act = plant_comp.massFlowRateActuator().get()
    mdot_req_act.setName('{}_Mdot_Req_Act'.format(plant_comp.nameString()))

    # heat pump COP curve
    constant_coeff = 1.932 + (cop - 3.65)
    hp_cop_curve = openstudio_model.CurveQuadratic(model)
    hp_cop_curve.setCoefficient1Constant(constant_coeff)
    hp_cop_curve.setCoefficient2x(0.227674286)
    hp_cop_curve.setCoefficient3xPOW2(-0.007313143)
    hp_cop_curve.setMinimumValueofx(1.67)
    hp_cop_curve.setMaximumValueofx(12.78)
    hp_cop_curve.setInputUnitTypeforX('Temperature')
    hp_cop_curve.setOutputUnitType('Dimensionless')

    # heat pump COP curve index variable
    hp_cop_curve_idx_var = \
        openstudio_model.EnergyManagementSystemCurveOrTableIndexVariable(model, hp_cop_curve)

    # high outlet temperature limit actuator
    tout_max_act = openstudio_model.EnergyManagementSystemActuator(
        plant_comp, 'Plant Connection 1', 'High Outlet Temperature Limit')
    tout_max_act.setName('{}_Tout_Max_Act'.format(plant_comp.nameString()))

    # init program
    init_pgrm = plant_comp.plantInitializationProgram().get()
    init_pgrm.setName('{}_Init_Pgrm'.format(plant_comp.nameString()))
    init_pgrm_body = \
        'SET Loop_Exit_Temp = {loop_exit_temp}\n' \
        'SET Loop_Delta_Temp = {design_temp_diff}\n' \
        'SET Cp = @CPHW Loop_Exit_Temp\n' \
        'SET rho = @RhoH2O Loop_Exit_Temp\n' \
        'SET {vdot_des_act_handle} = {vdot_des_int_var_handle}\n' \
        'SET {mdot_min_act_handle} = 0\n' \
        'SET Mdot_Max = {vdot_des_int_var_handle} * rho\n' \
        'SET {mdot_max_act_handle} = Mdot_Max\n' \
        'SET Cap = Mdot_Max * Cp * Loop_Delta_Temp\n' \
        'SET {cap_min_act_handle} = 0\n' \
        'SET {cap_max_act_handle} = Cap\n' \
        'SET {cap_opt_act_handle} = 1 * Cap'.format(
            loop_exit_temp=hot_water_loop.sizingPlant().designLoopExitTemperature(),
            design_temp_diff=hot_water_loop.sizingPlant().loopDesignTemperatureDifference(),
            vdot_des_act_handle=vdot_des_act.handle(),
            vdot_des_int_var_handle=vdot_des_int_var.handle(),
            mdot_min_act_handle=mdot_min_act.handle(),
            mdot_max_act_handle=mdot_max_act.handle(),
            cap_min_act_handle=cap_min_act.handle(),
            cap_max_act_handle=cap_max_act.handle(),
            cap_opt_act_handle=cap_opt_act.handle()
        )
    init_pgrm.setBody(init_pgrm_body)

    # sim program
    sim_pgrm = plant_comp.plantSimulationProgram().get()
    sim_pgrm.setName('{}_Sim_Pgrm'.format(plant_comp.nameString()))
    sim_pgrm_body = \
        'SET tmp = {load_int_var_handle}\n' \
        'SET tmp = {tin_int_var_handle}\n' \
        'SET tmp = {mdot_int_var_handle}\n' \
        'SET {tout_max_act_handle} = 75.0\n' \
        'IF {load_int_var_handle} == 0\n' \
        'SET {tout_act_handle} = {tin_int_var_handle}\n' \
        'SET {mdot_req_act_handle} = 0\n' \
        'SET Elec = 0\n' \
        'RETURN\n' \
        'ENDIF\n' \
        'IF {load_int_var_handle} >= {cap_max_act_handle}\n' \
        'SET Qdot = {cap_max_act_handle}\n' \
        'SET Mdot = {mdot_max_act_handle}\n' \
        'SET {mdot_req_act_handle} = Mdot\n' \
        'SET {tout_act_handle} = (Qdot / (Mdot * {cp_int_var_handle})) + {tin_int_var_handle}\n' \
        'IF {tout_act_handle} > {tout_max_act_handle}\n' \
        'SET {tout_act_handle} = {tout_max_act_handle}\n' \
        'SET Qdot = Mdot * {cp_int_var_handle} * ({tout_act_handle} - {tin_int_var_handle})\n' \
        'ENDIF\n' \
        'ELSE\n' \
        'SET Qdot = {load_int_var_handle}\n' \
        'SET {tout_act_handle} = {setpt_mgr_sch_sen_handle}\n' \
        'SET Mdot = Qdot / ({cp_int_var_handle} * ({tout_act_handle} - {tin_int_var_handle}))\n' \
        'SET {mdot_req_act_handle} = Mdot\n' \
        'ENDIF\n' \
        'SET Tdb = {oa_dbt_sen_handle}\n' \
        'SET COP = @CurveValue {hp_cop_curve_idx_var_handle} Tdb\n' \
        'SET EIR = 1 / COP\n' \
        'SET Pwr = Qdot * EIR\n' \
        'SET Elec = Pwr * SystemTimestep * 3600'.format(
            load_int_var_handle=load_int_var.handle(),
            tin_int_var_handle=tin_int_var.handle(),
            mdot_int_var_handle=mdot_int_var.handle(),
            tout_max_act_handle=tout_max_act.handle(),
            tout_act_handle=tout_act.handle(),
            mdot_req_act_handle=mdot_req_act.handle(),
            cap_max_act_handle=cap_max_act.handle(),
            mdot_max_act_handle=mdot_max_act.handle(),
            cp_int_var_handle=cp_int_var.handle(),
            setpt_mgr_sch_sen_handle=setpt_mgr_sch_sen.handle(),
            oa_dbt_sen_handle=oa_dbt_sen.handle(),
            hp_cop_curve_idx_var_handle=hp_cop_curve_idx_var.handle()
        )
    sim_pgrm.setBody(sim_pgrm_body)

    # init program calling manager
    init_mgr = plant_comp.plantInitializationProgramCallingManager().get()
    init_mgr.setName('{}_Init_Pgrm_Mgr'.format(plant_comp.nameString()))

    # sim program calling manager
    sim_mgr = plant_comp.plantSimulationProgramCallingManager().get()
    sim_mgr.setName('{}_Sim_Pgrm_Mgr'.format(plant_comp.nameString()))

    # metered output variable
    elec_mtr_out_var = openstudio_model.EnergyManagementSystemMeteredOutputVariable(
        model, '{} Electricity Consumption'.format(plant_comp.nameString()))
    elec_mtr_out_var.setName('{} Electricity Consumption'.format(plant_comp.nameString()))
    elec_mtr_out_var.setEMSVariableName('Elec')
    elec_mtr_out_var.setUpdateFrequency('SystemTimestep')
    elec_mtr_out_var.setString(4, str(sim_pgrm.handle()))
    elec_mtr_out_var.setResourceType('Electricity')
    elec_mtr_out_var.setGroupType('HVAC')
    elec_mtr_out_var.setEndUseCategory('Heating')
    elec_mtr_out_var.setEndUseSubcategory('')
    elec_mtr_out_var.setUnits('J')

    # add to supply side of hot water loop if specified
    if hot_water_loop is not None:
        hot_water_loop.addSupplyBranchForComponent(plant_comp)

    # add operation scheme
    htg_op_scheme = openstudio_model.PlantEquipmentOperationHeatingLoad(model)
    htg_op_scheme.addEquipment(1000000000, plant_comp)
    hot_water_loop.setPlantEquipmentOperationHeatingLoad(htg_op_scheme)

    return plant_comp



"""honeybee-openstudio translation commands."""
import sys
import logging
import click

from ladybug.commandutil import process_content_to_output
from honeybee.model import Model

import openstudio
from honeybee_openstudio.writer import model_to_openstudio

_logger = logging.getLogger(__name__)


@click.group(help='Commands for translating between Honeybee and OpenStudio.')
def translate():
    pass


@translate.command('model-to-osm')
@click.argument('model-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--output-file', '-o', help='Optional OSM file path to output the OSM string '
    'of the translation. By default this will be printed to stdout.',
    type=click.File('w'), default='-', show_default=True)
def model_to_osm_cli(model_file, output_file):
    """Translate a Honeybee Model to an OSM file.

    \b
    Args:
        model_file: Full path to a Honeybee Model file (HBJSON or HBpkl).
    """
    try:
        model_to_osm(model_file, output_file)
    except Exception as e:
        _logger.exception(f'Model translation failed:\n{e}')
        sys.exit(1)
    else:
        sys.exit(0)


def model_to_osm(model_file, output_file=None):
    """Translate a Honeybee Model to an OSM file.

    Args:
        model_file: Full path to a Honeybee Model file (HBJSON or HBpkl).
        output_file: Optional OSM file path to output the OSM string of the
            translation. If None, the string will be returned from this function.
    """
    model = Model.from_file(model_file)
    os_model = model_to_openstudio(model, print_progress=True)
    if output_file is not None and 'stdout' not in str(output_file):
        output_file = output_file.name \
            if not isinstance(output_file, str) else output_file
        os_model.save(output_file, overwrite=True)
    else:
        output = process_content_to_output(str(os_model), output_file)
        return output


@translate.command('osm-to-idf')
@click.argument('osm-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--output-file', '-o', help='Optional IDF file path to output the IDF string '
    'of the translation. By default this will be printed to stdout.',
    type=click.File('w'), default='-', show_default=True)
def osm_to_idf_cli(osm_file, output_file):
    """Translate a OSM file to an IDF file.

    \b
    Args:
        osm_file: Full path to a OpenStudio Model file (OSM).
    """
    try:
        osm_to_idf(osm_file, output_file)
    except Exception as e:
        _logger.exception(f'Model translation failed:\n{e}')
        sys.exit(1)
    else:
        sys.exit(0)


def osm_to_idf(osm_file, output_file=None):
    """Translate a Honeybee Model to an OSM file.

    Args:
        osm_file: Full path to an OpenStudio Model file (OSM).
        output_file: Optional IDF file path to output the IDF string of the
            translation. If None, the string will be returned from this function.
    """
    # load the model object from the OSM file
    exist_os_model = openstudio.model.Model.load(osm_file)
    if exist_os_model.is_initialized():
        os_model = exist_os_model.get()
    else:
        raise ValueError(
            'The file at "{}" does not appear to be an OpenStudio model.'.format(
                osm_file
            ))

    # translate the OpenStudio model to an IDF file
    idf_translator = openstudio.energyplus.ForwardTranslator()
    workspace = idf_translator.translateModel(os_model)

    # write the IDF file
    if output_file is not None and 'stdout' not in str(output_file):
        output_file = output_file.name \
            if not isinstance(output_file, str) else output_file
        workspace.save(output_file, overwrite=True)
    else:
        output = process_content_to_output(str(workspace), output_file)
        return output


@translate.command('append-to-osm')
@click.argument('osm-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('model-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--output-file', '-o', help='Optional OSM file path to output the OSM string '
    'of the translation. By default this will be printed to stdout.',
    type=click.File('w'), default='-', show_default=True)
def append_to_osm_cli(osm_file, model_file, output_file):
    """Append a Honeybee Model to a OSM file.

    \b
    Args:
        osm_file: Full path to a OpenStudio Model file (OSM).
        model_file: Full path to a Honeybee Model file (HBJSON or HBpkl).
    """
    try:
        append_to_osm(osm_file, model_file, output_file)
    except Exception as e:
        _logger.exception(f'Model appending failed:\n{e}')
        sys.exit(1)
    else:
        sys.exit(0)


def append_to_osm(osm_file, model_file, output_file=None):
    """Append a Honeybee Model to a OSM file.

    Args:
        osm_file: Full path to an OpenStudio Model file (OSM).
        model_file: Full path to a Honeybee Model file (HBJSON or HBpkl).
        output_file: Optional IDF file path to output the IDF string of the
            translation. If None, the string will be returned from this function.
    """
    # load the model object from the OSM file
    exist_os_model = openstudio.model.Model.load(osm_file)
    if exist_os_model.is_initialized():
        os_model = exist_os_model.get()
    else:
        raise ValueError(
            'The file at "{}" does not appear to be an OpenStudio model.'.format(
                osm_file
            ))

    # load the honeybee Model object
    model = Model.from_file(model_file)
    # append the honeybee model to the OSM
    os_model = model_to_openstudio(model, os_model)

    # write the OSM file
    if output_file is not None and 'stdout' not in str(output_file):
        output_file = output_file.name \
            if not isinstance(output_file, str) else output_file
        os_model.save(output_file, overwrite=True)
    else:
        output = process_content_to_output(str(os_model), output_file)
        return output



# coding=utf-8
"""OpenStudio translators for template HVAC systems."""
from __future__ import division

from honeybee.typing import clean_ep_string
from honeybee_energy.hvac.allair._base import _AllAirBase
from honeybee_energy.hvac.doas._base import _DOASBase
from honeybee_energy.hvac.allair.ptac import PTAC
from honeybee_energy.hvac.allair.psz import PSZ
from honeybee_energy.hvac.allair.pvav import PVAV
from honeybee_energy.hvac.allair.vav import VAV
from honeybee_energy.hvac.allair.furnace import ForcedAirFurnace
from honeybee_energy.hvac.doas.fcu import FCUwithDOAS
from honeybee_energy.hvac.doas.vrf import VRFwithDOAS
from honeybee_energy.hvac.doas.wshp import WSHPwithDOAS
from honeybee_energy.hvac.doas.radiant import RadiantwithDOAS
from honeybee_energy.hvac.heatcool.baseboard import Baseboard
from honeybee_energy.hvac.heatcool.evapcool import EvaporativeCooler
from honeybee_energy.hvac.heatcool.fcu import FCU
from honeybee_energy.hvac.heatcool.gasunit import GasUnitHeater
from honeybee_energy.hvac.heatcool.residential import Residential
from honeybee_energy.hvac.heatcool.vrf import VRF
from honeybee_energy.hvac.heatcool.windowac import WindowAC
from honeybee_energy.hvac.heatcool.wshp import WSHP
from honeybee_energy.hvac.heatcool.radiant import Radiant

from honeybee_openstudio.openstudio import openstudio, openstudio_model, \
    OSScheduleRuleset, OSTime
from .standards.hvac_systems import model_add_hvac_system, model_add_low_temp_radiant, \
    model_get_or_add_chilled_water_loop, model_add_hw_loop, model_add_chw_loop, \
    model_add_cw_loop, model_get_or_add_ambient_water_loop


def template_hvac_to_openstudio(hvac, os_zones, os_model):
    """Convert Honeybee HVAC TemplateSystem to OpenStudio.

    Args:
        hvac: Any honeybee-energy TemplateSystem class instance to be translated
            to OpenStudio.
        os_zones: A dictionary with two keys, each of which has a value for a
            list of OpenStudio ThermalZones. The keys are heated_zones and
            cooled_zones and the lists under each key note the OpenStudio
            ThermalZones to be given heating and cooling equipment by the HVAC.
        os_model: The OpenStudio Model object to which the HVAC system
            will be added.
    """
    # unpack the heated and cooled zones and organize them into groups
    heated_zones, cooled_zones = os_zones['heated_zones'], os_zones['cooled_zones']
    heated_and_cooled_zones, cooled_only_zones, heated_only_zones = [], [], []
    heat_dict = {z.nameString(): z for z in heated_zones}
    cool_dict = {z.nameString(): z for z in cooled_zones}
    all_dict = heat_dict.copy()
    all_dict.update(cool_dict)
    zones = list(all_dict.values())
    for zone in zones:
        zone_name = zone.nameString()
        if zone_name in heat_dict and zone_name in cool_dict:
            heated_and_cooled_zones.append(zone)
        elif zone_name in cool_dict:
            cooled_only_zones.append(zone)
        else:
            heated_only_zones.append(zone)
    system_zones = heated_and_cooled_zones + cooled_only_zones

    # determine the DOAS type from the demand controlled ventilation
    dcv = getattr(hvac, 'demand_controlled_ventilation', False)
    doas_type = 'DOAS' if not dcv else 'DOAS with DCV'
    air_loop = None  # will be returned from HVAC creation

    # add the HVAC system equipment using the equipment type
    # system type naming convention:
    # [ventilation strategy] [cooling system and plant] [heating system and plant]
    equip = hvac.equipment_type
    if isinstance(hvac, Baseboard):
        if equip == 'ElectricBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'BoilerBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'ASHPBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_zones)

        elif equip == 'DHWBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_zones)

    elif isinstance(hvac, EvaporativeCooler):
        if equip == 'EvapCoolers_ElectricBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)
            model_add_hvac_system(os_model, 'Evaporative Cooler',
                                  None, None, 'Electricity', cooled_zones)

        elif equip == 'EvapCoolers_BoilerBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)
            model_add_hvac_system(os_model, 'Evaporative Cooler', None, None,
                                  'Electricity', cooled_zones)

        elif equip == 'EvapCoolers_ASHPBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_zones)
            model_add_hvac_system(os_model, 'Evaporative Cooler', None, None,
                                  'Electricity', cooled_zones)

        elif equip == 'EvapCoolers_DHWBaseboard':
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_zones)
            model_add_hvac_system(os_model, 'Evaporative Cooler', None, None,
                                  'Electricity', cooled_zones)

        elif equip == 'EvapCoolers_Furnace':
            # use unit heater to represent forced air furnace
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)
            model_add_hvac_system(os_model, 'Evaporative Cooler', None, None,
                                  'Electricity', cooled_zones)

        elif equip == 'EvapCoolers_UnitHeaters':
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)
            model_add_hvac_system(os_model, 'Evaporative Cooler', None, None,
                                  'Electricity', cooled_zones)

        elif equip == 'EvapCoolers':
            model_add_hvac_system(os_model, 'Evaporative Cooler', None, None,
                                  'Electricity', cooled_zones)

    elif isinstance(hvac, FCUwithDOAS):
        if equip == 'DOAS_FCU_Chiller_Boiler':
            air_loop = model_add_hvac_system(os_model, doas_type, 'NaturalGas', None,
                                             'Electricity', zones)
            model_add_hvac_system(os_model, 'Fan Coil', 'NaturalGas', None, 'Electricity',
                                  zones, zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_Chiller_ASHP':
            air_loop = model_add_hvac_system(os_model, doas_type, 'AirSourceHeatPump',
                                             None, 'Electricity', zones)
            model_add_hvac_system(os_model, 'Fan Coil', 'AirSourceHeatPump', None,
                                  'Electricity', zones, zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_Chiller_DHW':
            air_loop = model_add_hvac_system(os_model, doas_type, 'DistrictHeating',
                                             None, 'Electricity', zones)
            model_add_hvac_system(os_model, 'Fan Coil', 'DistrictHeating', None,
                                  'Electricity', zones, zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_Chiller_ElectricBaseboard':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'Electricity', zones)
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  zone_equipment_ventilation=False)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity', None,
                                  None, heated_zones)

        elif equip == 'DOAS_FCU_Chiller_GasHeaters':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'Electricity', zones)
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  zone_equipment_ventilation=False)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas', None,
                                  None, heated_zones)

        elif equip == 'DOAS_FCU_Chiller':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'Electricity', zones)
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_ACChiller_Boiler':
            air_loop = model_add_hvac_system(os_model, doas_type, 'NaturalGas', None,
                                             'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Fan Coil', 'NaturalGas', None, 'Electricity',
                                  zones, chilled_water_loop_cooling_type='AirCooled',
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_ACChiller_ASHP':
            air_loop = model_add_hvac_system(os_model, doas_type, 'AirSourceHeatPump', None,
                                             'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Fan Coil', 'AirSourceHeatPump', None,
                                  'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled',
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_ACChiller_DHW':
            air_loop = model_add_hvac_system(os_model, doas_type, 'DistrictHeating', None,
                                             'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Fan Coil', 'DistrictHeating', None,
                                  'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled',
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_ACChiller_ElectricBaseboard':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled',
                                  zone_equipment_ventilation=False)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'DOAS_FCU_ACChiller_GasHeaters':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled',
                                  zone_equipment_ventilation=False)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas', None,
                                  None, heated_zones)

        elif equip == 'DOAS_FCU_ACChiller':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled',
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_DCW_Boiler':
            air_loop = model_add_hvac_system(os_model, doas_type, 'NaturalGas', None,
                                             'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Fan Coil', 'NaturalGas', None,
                                  'DistrictCooling', zones,
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_DCW_ASHP':
            air_loop = model_add_hvac_system(os_model, doas_type, 'AirSourceHeatPump',
                                             None, 'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Fan Coil', 'AirSourceHeatPump',
                                  None, 'DistrictCooling', zones,
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_DCW_DHW':
            air_loop = model_add_hvac_system(os_model, doas_type, 'DistrictHeating',
                                             None, 'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Fan Coil', 'DistrictHeating',
                                  None, 'DistrictCooling', zones,
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_FCU_DCW_ElectricBaseboard':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'DistrictCooling',
                                  zones, zone_equipment_ventilation=False)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'DOAS_FCU_DCW_GasHeaters':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'DistrictCooling',
                                  zones, zone_equipment_ventilation=False)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'DOAS_FCU_DCW':
            air_loop = model_add_hvac_system(os_model, doas_type, None, None,
                                             'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'DistrictCooling',
                                  zones, zone_equipment_ventilation=False)

    elif isinstance(hvac, RadiantwithDOAS):
        chilled_water_loop_cooling_type = 'WaterCooled'
        if equip == 'DOAS_Radiant_Chiller_Boiler':
            main_heat_fuel, cool_fuel = 'NaturalGas', 'Electricity'
        elif equip == 'DOAS_Radiant_Chiller_ASHP':
            main_heat_fuel, cool_fuel = 'AirSourceHeatPump', 'Electricity'
        elif equip == 'DOAS_Radiant_Chiller_DHW':
            main_heat_fuel, cool_fuel = 'DistrictHeating', 'Electricity'
        elif equip == 'DOAS_Radiant_ACChiller_Boiler':
            main_heat_fuel, cool_fuel = 'NaturalGas', 'Electricity'
            chilled_water_loop_cooling_type = 'AirCooled'
        elif equip == 'DOAS_Radiant_ACChiller_ASHP':
            main_heat_fuel, cool_fuel = 'AirSourceHeatPump', 'Electricity'
            chilled_water_loop_cooling_type = 'AirCooled'
        elif equip == 'DOAS_Radiant_ACChiller_DHW':
            main_heat_fuel, cool_fuel = 'DistrictHeating', 'Electricity'
            chilled_water_loop_cooling_type = 'AirCooled'
        elif equip == 'DOAS_Radiant_DCW_Boiler':
            main_heat_fuel, cool_fuel = 'NaturalGas', 'DistrictCooling'
        elif equip == 'DOAS_Radiant_DCW_ASHP':
            main_heat_fuel, cool_fuel = 'AirSourceHeatPump', 'DistrictCooling'
        elif equip == 'DOAS_Radiant_DCW_DHW':
            main_heat_fuel, cool_fuel = 'DistrictHeating', 'DistrictCooling'

        air_loop = model_add_hvac_system(
            os_model, doas_type, main_heat_fuel, None, cool_fuel, zones,
            chilled_water_loop_cooling_type=chilled_water_loop_cooling_type)
        hw_name = 'Low Temp Hot Water Loop'
        if os_model.getPlantLoopByName(hw_name).is_initialized():
            hot_water_loop = os_model.getPlantLoopByName(hw_name).get()
        else:
            hot_water_loop = model_add_hw_loop(
                os_model, main_heat_fuel, dsgn_sup_wtr_temp=120.0,
                boiler_draft_type='Condensing')
            hot_water_loop.setName(hw_name)
        chw_name = 'Low Temp Chilled Water Loop'
        if os_model.getPlantLoopByName(chw_name).is_initialized():
            chilled_water_loop = os_model.getPlantLoopByName(chw_name).get()
        else:
            if cool_fuel == 'DistrictCooling':
                chilled_water_loop = model_add_chw_loop(
                    os_model, chw_pumping_type='const_pri', cooling_fuel=cool_fuel)
            elif cool_fuel == 'HeatPump':
                condenser_water_loop = model_get_or_add_ambient_water_loop(os_model)
                chilled_water_loop = model_add_chw_loop(
                    os_model, chw_pumping_type='const_pri_var_sec',
                    chiller_cooling_type='WaterCooled',
                    chiller_compressor_type='Rotary Screw',
                    condenser_water_loop=condenser_water_loop)
            elif cool_fuel == 'Electricity':
                if chilled_water_loop_cooling_type == 'AirCooled':
                    chilled_water_loop = model_add_chw_loop(
                        os_model, chw_pumping_type='const_pri',
                        chiller_cooling_type='AirCooled', cooling_fuel=cool_fuel)
                else:
                    cond_name = 'Condenser Water Loop'
                    if os_model.getPlantLoopByName(cond_name).is_initialized():
                        condenser_water_loop = os_model.getPlantLoopByName(cond_name).get()
                    else:
                        fan_type = 'Variable Speed Fan'
                        condenser_water_loop = model_add_cw_loop(
                            os_model, cooling_tower_type='Open Cooling Tower',
                            cooling_tower_fan_type='Propeller or Axial',
                            cooling_tower_capacity_control=fan_type,
                            number_of_cells_per_tower=1, number_cooling_towers=1)
                        condenser_water_loop.setName(cond_name)
                    chilled_water_loop = model_add_chw_loop(
                        os_model, chw_pumping_type='const_pri_var_sec',
                        chiller_cooling_type='WaterCooled',
                        chiller_compressor_type='Rotary Screw',
                        condenser_water_loop=condenser_water_loop)
            chilled_water_loop.setName(chw_name)

        control_strategy, include_carpet = 'proportional_control', False
        radiant_temperature_control_type = 'SurfaceFaceTemperature'
        if hvac.radiant_type in ('CeilingMetalPanel', 'FloorWithHardwood'):
            control_strategy = 'none'
            radiant_temperature_control_type = 'OperativeTemperature'
        radiant_type = hvac.radiant_type.lower()
        if hvac.radiant_type == 'FloorWithCarpet':
            radiant_type, include_carpet = 'floor', True

        model_add_low_temp_radiant(
            os_model, zones, hot_water_loop, chilled_water_loop, radiant_type=radiant_type,
            include_carpet=include_carpet, control_strategy=control_strategy,
            radiant_temperature_control_type=radiant_temperature_control_type,
            radiant_availability_type='all_day')

    elif isinstance(hvac, VRFwithDOAS):
        if equip == 'DOAS_VRF':
            air_loop = model_add_hvac_system(os_model, doas_type, 'Electricity',
                                             None, 'Electricity', zones,
                                             air_loop_heating_type='DX',
                                             air_loop_cooling_type='DX')
            model_add_hvac_system(os_model, 'VRF', 'Electricity',
                                  None, 'Electricity', zones)

    elif isinstance(hvac, WSHPwithDOAS):
        if equip == 'DOAS_WSHP_FluidCooler_Boiler':
            air_loop = model_add_hvac_system(os_model, doas_type, 'NaturalGas', None,
                                             'Electricity', zones)
            model_add_hvac_system(os_model, 'Water Source Heat Pumps', 'NaturalGas',
                                  None, 'Electricity', zones,
                                  heat_pump_loop_cooling_type='FluidCooler',
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_WSHP_CoolingTower_Boiler':
            air_loop = model_add_hvac_system(os_model, doas_type, 'NaturalGas', None,
                                             'Electricity', zones)
            model_add_hvac_system(os_model, 'Water Source Heat Pumps', 'NaturalGas',
                                  None, 'Electricity', zones,
                                  heat_pump_loop_cooling_type='CoolingTower',
                                  zone_equipment_ventilation=False)

        elif equip == 'DOAS_WSHP_GSHP':
            air_loop = model_add_hvac_system(os_model, doas_type, 'Electricity', None,
                                             'Electricity', zones, air_loop_heating_type='DX',
                                             air_loop_cooling_type='DX')
            model_add_hvac_system(os_model, 'Ground Source Heat Pumps', 'Electricity', None,
                                  'Electricity', zones, zone_equipment_ventilation=False)

        elif equip == 'DOAS_WSHP_DCW_DHW':
            air_loop = model_add_hvac_system(os_model, doas_type, 'DistrictHeating', None,
                                             'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Water Source Heat Pumps', 'DistrictHeating',
                                  None, 'DistrictCooling', zones,
                                  zone_equipment_ventilation=False)

    # ventilation provided by zone fan coil unit in fan coil systems
    elif isinstance(hvac, FCU):
        if equip == 'FCU_Chiller_Boiler':
            model_add_hvac_system(os_model, 'Fan Coil', 'NaturalGas', None,
                                  'Electricity', zones)

        elif equip == 'FCU_Chiller_ASHP':
            model_add_hvac_system(os_model, 'Fan Coil', 'AirSourceHeatPump',
                                  None, 'Electricity', zones)

        elif equip == 'FCU_Chiller_DHW':
            model_add_hvac_system(os_model, 'Fan Coil', 'DistrictHeating',
                                  None, 'Electricity', zones)

        elif equip == 'FCU_Chiller_ElectricBaseboard':
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'FCU_Chiller_GasHeaters':
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'FCU_Chiller':
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones)

        elif equip == 'FCU_ACChiller_Boiler':
            model_add_hvac_system(os_model, 'Fan Coil', 'NaturalGas', None, 'Electricity',
                                  zones, chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'FCU_ACChiller_ASHP':
            model_add_hvac_system(os_model, 'Fan Coil', 'AirSourceHeatPump',
                                  None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'FCU_ACChiller_DHW':
            model_add_hvac_system(os_model, 'Fan Coil', 'DistrictHeating', None,
                                  'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'FCU_ACChiller_ElectricBaseboard':
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'FCU_ACChiller_GasHeaters':
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled')
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'FCU_ACChiller':
            model_add_hvac_system(os_model, 'Fan Coil', None, None, 'Electricity', zones,
                                  chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'FCU_DCW_Boiler':
            model_add_hvac_system(os_model, 'Fan Coil ', 'NaturalGas', None,
                                  'DistrictCooling', zones)

        elif equip == 'FCU_DCW_ASHP':
            model_add_hvac_system(os_model, 'Fan Coil', 'AirSourceHeatPump',
                                  None, 'DistrictCooling', zones)

        elif equip == 'FCU_DCW_DHW':
            model_add_hvac_system(os_model, 'Fan Coil', 'DistrictHeating',
                                  None, 'DistrictCooling', zones)

        elif equip == 'FCU_DCW_ElectricBaseboard':
            model_add_hvac_system(os_model, 'Fan Coil', None, None,
                                  'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'FCU_DCW_GasHeaters':
            model_add_hvac_system(os_model, 'Fan Coil', None, None,
                                  'DistrictCooling', zones)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'FCU_DCW':
            model_add_hvac_system(os_model, 'Fan Coil', None, None,
                                  'DistrictCooling', zones)

    elif isinstance(hvac, Radiant):
        chilled_water_loop_cooling_type = 'WaterCooled'
        if equip == 'Radiant_Chiller_Boiler':
            main_heat_fuel, cool_fuel = 'NaturalGas', 'Electricity'
        elif equip == 'Radiant_Chiller_ASHP':
            main_heat_fuel, cool_fuel = 'AirSourceHeatPump', 'Electricity'
        elif equip == 'Radiant_Chiller_DHW':
            main_heat_fuel, cool_fuel = 'DistrictHeating', 'Electricity'
        elif equip == 'Radiant_ACChiller_Boiler':
            main_heat_fuel, cool_fuel = 'NaturalGas', 'Electricity'
            chilled_water_loop_cooling_type = 'AirCooled'
        elif equip == 'Radiant_ACChiller_ASHP':
            main_heat_fuel, cool_fuel = 'AirSourceHeatPump', 'Electricity'
            chilled_water_loop_cooling_type = 'AirCooled'
        elif equip == 'Radiant_ACChiller_DHW':
            main_heat_fuel, cool_fuel = 'DistrictHeating', 'Electricity'
            chilled_water_loop_cooling_type = 'AirCooled'
        elif equip == 'Radiant_DCW_Boiler':
            main_heat_fuel, cool_fuel = 'NaturalGas', 'DistrictCooling'
        elif equip == 'Radiant_DCW_ASHP':
            main_heat_fuel, cool_fuel = 'AirSourceHeatPump', 'DistrictCooling'
        elif equip == 'Radiant_DCW_DHW':
            main_heat_fuel, cool_fuel = 'DistrictHeating', 'DistrictCooling'

        hw_name = 'Low Temperature Hot Water Loop'
        if os_model.getPlantLoopByName(hw_name).is_initialized():
            hot_water_loop = os_model.getPlantLoopByName(hw_name).get()
        else:
            hot_water_loop = model_add_hw_loop(
                os_model, main_heat_fuel, dsgn_sup_wtr_temp=120.0,
                boiler_draft_type='Condensing')
            hot_water_loop.setName(hw_name)
        chilled_water_loop = model_get_or_add_chilled_water_loop(
            os_model, cool_fuel,
            chilled_water_loop_cooling_type=chilled_water_loop_cooling_type)

        control_strategy, include_carpet = 'proportional_control', False
        radiant_temperature_control_type = 'SurfaceFaceTemperature'
        if hvac.radiant_type in ('CeilingMetalPanel', 'FloorWithHardwood'):
            control_strategy = 'none'
            radiant_temperature_control_type = 'OperativeTemperature'
        radiant_type = hvac.radiant_type.lower()
        if hvac.radiant_type == 'FloorWithCarpet':
            radiant_type, include_carpet = 'floor', True

        model_add_low_temp_radiant(
            os_model, zones, hot_water_loop, chilled_water_loop, radiant_type=radiant_type,
            include_carpet=include_carpet, control_strategy=control_strategy,
            radiant_temperature_control_type=radiant_temperature_control_type,
            radiant_availability_type='all_day')

    elif isinstance(hvac, ForcedAirFurnace):
        if equip == 'Furnace':
            # includes ventilation, whereas residential forced air furnace does not.
            model_add_hvac_system(os_model, 'Forced Air Furnace', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'Furnace_Electric':
            # includes ventilation, whereas residential forced air furnace does not.
            model_add_hvac_system(os_model, 'Forced Air Furnace', 'Electricity',
                                  None, None, heated_zones)

    elif isinstance(hvac, GasUnitHeater):
        if equip == 'GasHeaters':
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

    elif isinstance(hvac, PTAC):
        if equip == 'PTAC_ElectricBaseboard':
            model_add_hvac_system(os_model, 'PTAC', None, None, 'Electricity',
                                  system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'PTAC_BoilerBaseboard':
            model_add_hvac_system(os_model, 'PTAC', None, None, 'Electricity',
                                  system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'PTAC_DHWBaseboard':
            model_add_hvac_system(os_model, 'PTAC', None, None, 'Electricity',
                                  system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_zones)

        elif equip == 'PTAC_GasHeaters':
            model_add_hvac_system(os_model, 'PTAC', None, None, 'Electricity',
                                  system_zones)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'PTAC_ElectricCoil':
            model_add_hvac_system(os_model, 'PTAC', None, 'Electricity',
                                  'Electricity', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

        elif equip == 'PTAC_GasCoil':
            model_add_hvac_system(os_model, 'PTAC', None, 'NaturalGas',
                                  'Electricity', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

        elif equip == 'PTAC_Boiler':
            model_add_hvac_system(os_model, 'PTAC', 'NaturalGas', None,
                                  'Electricity', system_zones)
            # use 'Baseboard gas boiler' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_only_zones)

        elif equip == 'PTAC_ASHP':
            model_add_hvac_system(os_model, 'PTAC', 'AirSourceHeatPump',
                                  None, 'Electricity', system_zones)
            # use 'Baseboard central air source heat pump' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_only_zones)

        elif equip == 'PTAC_DHW':
            model_add_hvac_system(os_model, 'PTAC', 'DistrictHeating', None,
                                  'Electricity', system_zones)
            # use 'Baseboard district hot water heat' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_only_zones)

        elif equip == 'PTAC':
            model_add_hvac_system(os_model, 'PTAC', None, None, 'Electricity',
                                  system_zones)

        elif equip == 'PTHP':
            model_add_hvac_system(os_model, 'PTHP', 'Electricity', None,
                                  'Electricity', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

    elif isinstance(hvac, PSZ):
        if equip == 'PSZAC_ElectricBaseboard':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'Electricity', system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_BoilerBaseboard':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'Electricity', system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_DHWBaseboard':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'Electricity', system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_GasHeaters':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'Electricity', system_zones)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_ElectricCoil':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, 'Electricity',
                                             'Electricity', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_GasCoil':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, 'NaturalGas',
                                             'Electricity', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_Boiler':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', 'NaturalGas', None,
                                             'Electricity', system_zones)
            # use 'Baseboard gas boiler' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_ASHP':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', 'AirSourceHeatPump',
                                             None, 'Electricity', system_zones)
            # use 'Baseboard central air source heat pump' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_DHW':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', 'DistrictHeating',
                                             None, 'Electricity', system_zones)
            # use 'Baseboard district hot water' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'Electricity', cooled_zones)

        elif equip == 'PSZAC_DCW_ElectricBaseboard':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'DistrictCooling', system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_DCW_BoilerBaseboard':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'DistrictCooling', system_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_DCW_GasHeaters':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'DistrictCooling', system_zones)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'PSZAC_DCW_ElectricCoil':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, 'Electricity',
                                             'DistrictCooling', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_DCW_GasCoil':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, 'NaturalGas',
                                             'DistrictCooling', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_DCW_Boiler':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', 'NaturalGas', None,
                                             'DistrictCooling', system_zones)
            # use 'Baseboard gas boiler' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_DCW_ASHP':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', 'AirSourceHeatPump',
                                             None, 'DistrictCooling', system_zones)
            # use 'Baseboard central air source heat pump' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_DCW_DHW':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', 'DistrictHeating',
                                             None, 'DistrictCooling', system_zones)
            # use 'Baseboard district hot water' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_only_zones)

        elif equip == 'PSZAC_DCW':
            air_loop = model_add_hvac_system(os_model, 'PSZ-AC', None, None,
                                             'DistrictCooling', cooled_zones)

        elif equip == 'PSZHP':
            air_loop = model_add_hvac_system(os_model, 'PSZ-HP', 'Electricity', None,
                                             'Electricity', system_zones)
            # use 'Baseboard electric' for heated only zones
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_only_zones)

    elif isinstance(hvac, PVAV):  # PVAV systems by default use a DX coil for cooling
        if equip == 'PVAV_Boiler':
            air_loop = model_add_hvac_system(os_model, 'PVAV Reheat', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones)

        elif equip == 'PVAV_ASHP':
            air_loop = model_add_hvac_system(os_model, 'PVAV Reheat', 'AirSourceHeatPump',
                                             'AirSourceHeatPump', 'Electricity', zones)

        elif equip == 'PVAV_DHW':
            air_loop = model_add_hvac_system(os_model, 'PVAV Reheat', 'DistrictHeating',
                                             'DistrictHeating', 'Electricity', zones)

        elif equip == 'PVAV_PFP':
            air_loop = model_add_hvac_system(os_model, 'PVAV PFP Boxes', 'Electricity',
                                             'Electricity', 'Electricity', zones)

        elif equip == 'PVAV_BoilerElectricReheat':
            air_loop = model_add_hvac_system(os_model, 'PVAV Reheat', 'Gas',
                                             'Electricity', 'Electricity', zones)

    elif isinstance(hvac, Residential):  # all residential systems have no ventilation
        if equip == 'ResidentialAC_ElectricBaseboard':
            model_add_hvac_system(os_model, 'Residential AC', None, None, None,
                                  cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'ResidentialAC_BoilerBaseboard':
            model_add_hvac_system(os_model, 'Residential AC', None, None, None,
                                  cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'ResidentialAC_ASHPBaseboard':
            model_add_hvac_system(os_model, 'Residential AC', None, None, None,
                                  cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_zones)

        elif equip == 'ResidentialAC_DHWBaseboard':
            model_add_hvac_system(os_model, 'Residential AC', None, None, None,
                                  cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_zones)

        elif equip == 'ResidentialAC_ResidentialFurnace':
            model_add_hvac_system(os_model, 'Residential Forced Air Furnace with AC',
                                  None, None, None, zones)

        elif equip == 'ResidentialAC':
            model_add_hvac_system(os_model, 'Residential AC', None, None, None,
                                  cooled_zones)

        elif equip == 'ResidentialHP':
            model_add_hvac_system(os_model, 'Residential Air Source Heat Pump',
                                  'Electricity', None, 'Electricity', zones)

        elif equip == 'ResidentialHPNoCool':
            model_add_hvac_system(os_model, 'Residential Air Source Heat Pump',
                                  'Electricity', None, None, heated_zones)

        elif equip == 'ResidentialFurnace':
            model_add_hvac_system(os_model, 'Residential Forced Air Furnace',
                                  'NaturalGas', None, None, zones)

    elif isinstance(hvac, VAV):
        if equip == 'VAV_Chiller_Boiler':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones)

        elif equip == 'VAV_Chiller_ASHP':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'AirSourceHeatPump',
                                             'AirSourceHeatPump', 'Electricity', zones)

        elif equip == 'VAV_Chiller_DHW':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'DistrictHeating',
                                             'DistrictHeating', 'Electricity', zones)

        elif equip == 'VAV_Chiller_PFP':
            air_loop = model_add_hvac_system(os_model, 'VAV PFP Boxes', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones)

        elif equip == 'VAV_Chiller_GasCoil':
            air_loop = model_add_hvac_system(os_model, 'VAV Gas Reheat', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones)

        elif equip == 'VAV_ACChiller_Boiler':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'VAV_ACChiller_ASHP':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'AirSourceHeatPump',
                                             'AirSourceHeatPump', 'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'VAV_ACChiller_DHW':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'DistrictHeating',
                                             'DistrictHeating', 'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'VAV_ACChiller_PFP':
            air_loop = model_add_hvac_system(os_model, 'VAV PFP Boxes', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'VAV_ACChiller_GasCoil':
            air_loop = model_add_hvac_system(os_model, 'VAV Gas Reheat', 'NaturalGas',
                                             'NaturalGas', 'Electricity', zones,
                                             chilled_water_loop_cooling_type='AirCooled')

        elif equip == 'VAV_DCW_Boiler':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'NaturalGas',
                                             'NaturalGas', 'DistrictCooling', zones)

        elif equip == 'VAV_DCW_ASHP':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'AirSourceHeatPump',
                                             'AirSourceHeatPump', 'DistrictCooling', zones)

        elif equip == 'VAV_DCW_DHW':
            air_loop = model_add_hvac_system(os_model, 'VAV Reheat', 'DistrictHeating',
                                             'DistrictHeating', 'DistrictCooling', zones)

        elif equip == 'VAV_DCW_PFP':
            air_loop = model_add_hvac_system(os_model, 'VAV PFP Boxes', 'NaturalGas',
                                             'NaturalGas', 'DistrictCooling', zones)

        elif equip == 'VAV_DCW_GasCoil':
            air_loop = model_add_hvac_system(os_model, 'VAV Gas Reheat', 'NaturalGas',
                                             'NaturalGas', 'DistrictCooling', zones)

    elif isinstance(hvac, VRF):
        if equip == 'VRF':
            model_add_hvac_system(os_model, 'VRF', 'Electricity', None,
                                  'Electricity', zones)

    elif isinstance(hvac, WSHP):
        if equip == 'WSHP_FluidCooler_Boiler':
            model_add_hvac_system(os_model, 'Water Source Heat Pumps',
                                  'NaturalGas', None, 'Electricity', zones,
                                  heat_pump_loop_cooling_type='FluidCooler')

        elif equip == 'WSHP_CoolingTower_Boiler':
            model_add_hvac_system(os_model, 'Water Source Heat Pumps',
                                  'NaturalGas', None, 'Electricity', zones,
                                  heat_pump_loop_cooling_type='CoolingTower')

        elif equip == 'WSHP_GSHP':
            model_add_hvac_system(os_model, 'Ground Source Heat Pumps',
                                  'Electricity', None, 'Electricity', zones)

        elif equip == 'WSHP_DCW_DHW':
            model_add_hvac_system(os_model, 'Water Source Heat Pumps', 'DistrictHeating',
                                  None, 'DistrictCooling', zones)

    elif isinstance(hvac, WindowAC):
        if equip == 'WindowAC_ElectricBaseboard':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'Electricity',
                                  None, None, heated_zones)

        elif equip == 'WindowAC_BoilerBaseboard':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'WindowAC_ASHPBaseboard':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'AirSourceHeatPump',
                                  None, None, heated_zones)

        elif equip == 'WindowAC_DHWBaseboard':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)
            model_add_hvac_system(os_model, 'Baseboards', 'DistrictHeating',
                                  None, None, heated_zones)

        elif equip == 'WindowAC_Furnace':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)
            model_add_hvac_system(os_model, 'Forced Air Furnace', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'WindowAC_GasHeaters':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)
            model_add_hvac_system(os_model, 'Unit Heaters', 'NaturalGas',
                                  None, None, heated_zones)

        elif equip == 'WindowAC':
            model_add_hvac_system(os_model, 'Window AC', None, None,
                                  'Electricity', cooled_zones)

    else:
        raise ValueError('HVAC system type "{}" not recognized'.format(equip))

    # assign all of the properties associated with the air loop
    if air_loop is not None:
        # name the air loop with the name the user specified for the HVAC
        clean_hvac_name = clean_ep_string(hvac.display_name)
        if not isinstance(air_loop, list):
            air_loop.setName(clean_hvac_name)
            os_air_loops = [air_loop]
        else:
            os_air_loops = air_loop
            for i, loop in enumerate(os_air_loops):
                loop.setName('{} {}'.format(clean_hvac_name, i))

        # have an always available schedule ready to use if there are no user controls
        always_avail_name = 'Building HVAC Always Available'
        opt_sch = os_model.getScheduleByName(always_avail_name)
        if opt_sch.is_initialized():
            always_avail = opt_sch.get()
        else:
            always_avail = OSScheduleRuleset(os_model)
            always_avail.setName(always_avail_name)
            def_day_sch = always_avail.defaultDaySchedule()
            def_day_sch.addValue(OSTime(0, 24, 0, 0), 1)

        # assign the properties that are specific to All-Air systems
        if isinstance(hvac, _AllAirBase):
            for os_air_loop in os_air_loops:
                # set the loop to always be available
                os_air_loop.setAvailabilitySchedule(always_avail)
                # assign the economizer
                oasys = os_air_loop.airLoopHVACOutdoorAirSystem()
                if oasys.is_initialized():
                    os_oasys = oasys.get()
                    oactrl = os_oasys.getControllerOutdoorAir()
                    oactrl.setEconomizerControlType(hvac.economizer_type)
                    # assign demand controlled ventilation
                    if hvac.demand_controlled_ventilation:
                        vent_ctrl = oactrl.controllerMechanicalVentilation()
                        vent_ctrl.setDemandControlledVentilationNoFail(True)
                        oactrl.resetMinimumFractionofOutdoorAirSchedule()

        # assign the properties that are specific to DOAS systems
        if isinstance(hvac, _DOASBase):
            avail_sch = None
            if hvac.doas_availability_schedule is not None:
                sch_id = hvac.doas_availability_schedule.identifier
                schedule = os_model.getScheduleByName(sch_id)
                if schedule.is_initialized():
                    avail_sch = schedule.get()
            avail_sch = always_avail if avail_sch is None else avail_sch
            for os_air_loop in os_air_loops:
                os_air_loop.setAvailabilitySchedule(avail_sch)

        # set the heat recovery if it is specified
        if hvac.sensible_heat_recovery != 0 or hvac.latent_heat_recovery != 0:
            for os_air_loop in os_air_loops:
                heat_ex = _get_or_add_heat_recovery(os_model, os_air_loop)
                # ratio of max to standard efficiency from OpenStudio Standards
                eff_sens = hvac.sensible_heat_recovery
                heat_ex.setSensibleEffectivenessat100CoolingAirFlow(eff_sens)
                heat_ex.setSensibleEffectivenessat100HeatingAirFlow(eff_sens)
                eff_lat = hvac.latent_heat_recovery
                heat_ex.setLatentEffectivenessat100CoolingAirFlow(eff_lat)
                heat_ex.setLatentEffectivenessat100HeatingAirFlow(eff_lat)
                if os_model.version() < openstudio.VersionString('3.8.0'):
                    heat_ex.setSensibleEffectivenessat75CoolingAirFlow(eff_sens)
                    heat_ex.setSensibleEffectivenessat75HeatingAirFlow(eff_sens)
                    heat_ex.setLatentEffectivenessat75CoolingAirFlow(eff_lat)
                    heat_ex.setLatentEffectivenessat75HeatingAirFlow(eff_lat)

        # assign electric humidifier if there's an air loop and zones have humidistat
        humidistat_exists = False
        for zone in zones:
            h_stat = zone.zoneControlHumidistat()
            if h_stat.is_initialized():
                humidistat_exists = True
                if isinstance(hvac, _DOASBase):
                    z_sizing = zone.sizingZone()
                    z_sizing.setDedicatedOutdoorAirSystemControlStrategy(
                        'NeutralDehumidifiedSupplyAir')
        if humidistat_exists:
            for os_air_loop in os_air_loops:
                _add_humidifier(os_model, os_air_loop)

        # set the outdoor air controller to respect room-level ventilation schedules
        oa_sch, oa_sch_name, = None, None
        for i, zone in enumerate(zones):
            oa_spec = zone.spaces()[0].designSpecificationOutdoorAir()
            if oa_spec.is_initialized():
                oa_spec = oa_spec.get()
                space_oa_sch = oa_spec.outdoorAirFlowRateFractionSchedule()
                if space_oa_sch.is_initialized():
                    space_oa_sch = space_oa_sch.get()
                    space_oa_sch_name = space_oa_sch.nameString()
                    if i == 0 or space_oa_sch_name == oa_sch_name:
                        oa_sch, oa_sch_name = space_oa_sch, space_oa_sch_name
                    else:  # different schedules across zones; just use constant max
                        oa_sch = None
        oa_sch = always_avail if oa_sch is None else oa_sch
        for os_air_loop in os_air_loops:
            oasys = os_air_loop.airLoopHVACOutdoorAirSystem()
            if oasys.is_initialized():
                os_oasys = oasys.get()
                oactrl = os_oasys.getControllerOutdoorAir()
                oactrl.resetMinimumFractionofOutdoorAirSchedule()
                oactrl.setMinimumOutdoorAirSchedule(oa_sch)

    # if the systems are PTAC and there is ventilation, ensure the system includes it
    if isinstance(hvac, PTAC):
        always_on = os_model.alwaysOnDiscreteSchedule()
        for zone in zones:
            # check if the space type has ventilation assigned to it
            out_air = zone.spaces()[0].designSpecificationOutdoorAir()
            if out_air.is_initialized():
                # get any ventilation schedules
                vent_sched = always_on
                out_air = out_air.get()
                air_sch = out_air.outdoorAirFlowRateFractionSchedule()
                if air_sch.is_initialized():
                    vent_sched = air_sch.get()
                # get the PTAC object
                ptac = None
                for equip in zone.equipment():
                    e_name = equip.nameString()
                    if 'PTAC' in e_name:
                        ptac = os_model.getZoneHVACPackagedTerminalAirConditioner(
                            equip.handle())
                    elif 'PTHP' in e_name:
                        ptac = os_model.getZoneHVACPackagedTerminalHeatPump(
                            equip.handle())
                # assign the schedule to the PTAC object
                if ptac is not None and ptac.is_initialized():
                    ptac = ptac.get()
                    ptac.setSupplyAirFanOperatingModeSchedule(vent_sched)


def _get_or_add_heat_recovery(os_model, os_air_loop):
    """Get an existing heat exchanger in an air loop or add one if it does not exist."""
    # get an existing heat energy recovery unit from an air loop
    for supply_comp in os_air_loop.oaComponents():
        if supply_comp.to_HeatExchangerAirToAirSensibleAndLatent().is_initialized():
            return supply_comp.to_HeatExchangerAirToAirSensibleAndLatent().get()

    # create a heat recovery unit with default zero efficiencies
    heat_ex = openstudio_model.HeatExchangerAirToAirSensibleAndLatent(os_model)
    heat_ex.setEconomizerLockout(False)
    heat_ex.setName('{}_Heat Recovery Unit'.format(os_air_loop.nameString()))

    # add the heat exchanger to the air loop
    outdoor_node = os_air_loop.reliefAirNode()
    if outdoor_node.is_initialized():
        os_outdoor_node = outdoor_node.get()
        heat_ex.addToNode(os_outdoor_node)
    return heat_ex


def _add_humidifier(os_model, os_air_loop):
    """Add a humidifier to an air loop so it can meet humidification setpoints."""
    # create an electric humidifier
    humidifier = openstudio_model.HumidifierSteamElectric(os_model)
    humidifier.setName('{}_Humidifier Unit'.format(os_air_loop.nameString()))
    humid_control = openstudio_model.SetpointManagerMultiZoneHumidityMinimum(os_model)
    humid_control.setName('{}_Humidifier Controller'.format(os_air_loop.nameString()))

    # add the humidifier to the air loop
    supply_node = os_air_loop.supplyOutletNode()
    humidifier.addToNode(supply_node)
    humid_control.addToNode(supply_node)
    return humidifier
